#include "hark-lib/node.h"
#include "hark-lib/network.h"
#include "hark-lib/input.h"

#include "hark-lib/node/Constant.h"
#include "hark-lib/node/InputStream.h"
#include "hark-lib/node/MultiFFT.h"
#include "hark-lib/node/FeatureRemover.h"
#include "hark-lib/node/Delta.h"
#include "hark-lib/node/MelFilterBank.h"
#include "hark-lib/node/MSLSExtraction.h"
#include "hark-lib/node/PreEmphasis.h"
#include "hark-lib/node/SaveFeatures.h"
#include "hark-lib/node/SaveWavePCM.h"
#include "hark-lib/node/SpeechRecognitionClient.h"
#include "hark-lib/node/SpectralMeanNormalizationIncremental.h"
#include "hark-lib/node/VADZC.h"
#include "hark-lib/node/AudioStreamFromMemory.h"
#include "hark-lib/node/PublishData.h"
#include "hark-lib/node/SubscribeData.h"

#include <stdlib.h>
#include <exception>
#include <iostream>
#include <fstream>
#include <random>

using namespace hark;

class Extracting: public hark::NetworkDef
{
public:
    void
    build(hark::Network &network, hark::DataSourceMap &input, hark::DataSinkMap &output)
    {
        auto &node_multi_fft = network.create<hark::Node, hark::node::MultiFFT>();
        auto &node_feature_remover = network.create<hark::Node, hark::node::FeatureRemover>();
        auto &node_delta = network.create<hark::Node, hark::node::Delta>();
        auto &node_mel_filter_bank = network.create<hark::Node, hark::node::MelFilterBank>();
        auto &node_msls_extraction = network.create<hark::Node, hark::node::MSLSExtraction>();
        auto &node_pre_emphasis = network.create<hark::Node, hark::node::PreEmphasis>();

        node_multi_fft
            << hark::Input("INPUT", input["CHANNEL_WAV"]);
            // << hark::Input("LENGTH", 512)
            // << hark::Input("WINDOW", "CONJ")
            // << hark::Input("WINDOW_LENGTH", 512);

        node_feature_remover
            << hark::Input("INPUT", node_delta["OUTPUT"])
            << hark::Input("SELECTOR", "40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81");

        node_mel_filter_bank
            << hark::Input("INPUT", node_pre_emphasis["OUTPUT"])
            // << hark::Input("LENGTH", 512)
            // << hark::Input("SAMPLING_RATE", 16000)
            // << hark::Input("CUTOFF", 8000)
            // << hark::Input("MIN_FREQUENCY", 63)
            // << hark::Input("MAX_FREQUENCY", 8000)
            << hark::Input("FBANK_COUNT", 40);

        node_msls_extraction
            << hark::Input("FBANK", node_mel_filter_bank["OUTPUT"])
            << hark::Input("SPECTRUM", node_pre_emphasis["OUTPUT"])
            << hark::Input("FBANK_COUNT", 40)
            << hark::Input("NORMALIZATION_MODE", "SPECTRAL")
            // << hark::Input("USE_LEGACY_MODE", true)
            // << hark::Input("USE_HTK_LIFTER", false)
            // << hark::Input("LIFTERING_COEF", 22)
            << hark::Input("USE_POWER", true);

        node_pre_emphasis
            << hark::Input("INPUT", node_multi_fft["OUTPUT"])
            // << hark::Input("LENGTH", 512)
            // << hark::Input("SAMPLING_RATE", 16000)
            // << hark::Input("PREEMCOEF", 0.97f)
            << hark::Input("INPUT_TYPE", "SPECTRUM");

        node_delta
            << hark::Input("INPUT", node_msls_extraction["OUTPUT"]);
            // << hark::Input("FBANK_COUNT", 41);

        output["OUTPUT"] << node_feature_remover["OUTPUT"];
    }
};

class Recording: public hark::NetworkDef
{
public:
    void
    build(hark::Network &network, hark::DataSourceMap &input, hark::DataSinkMap &output)
    {
        auto &node_audio_stream_from_memory = network.create<hark::Node, hark::node::AudioStreamFromMemory, hark::dispatch::TriggeredMultiShot>();
        //auto &node_extracting = network.create<hark::Network, Extracting>("Extracting");

        auto &node_extracting = network.create_network(new Extracting(), new hark::dispatch::TriggeredRepeat(), "Extracting");

        auto &node_spectral_mean_normalization_incremental = network.create<hark::Node, hark::node::SpectralMeanNormalizationIncremental>();
        auto &node_vad_zc = network.create<hark::Node, hark::node::VADZC>();

        node_audio_stream_from_memory
            << hark::Input("INPUT", input["INPUT"]);
            // << hark::Input("LENGTH", 512)
            // << hark::Input("ADVANCE", 160)
            // << hark::Input("USE_WAIT", true);

        node_vad_zc
            << hark::Input("INPUT", node_audio_stream_from_memory["AUDIO"])
            // << hark::Input("ADVANCE", 160)
            // << hark::Input("LENGTH", 512)
            << hark::Input("CHANNEL_COUNT", 1)
            // << hark::Input("SAMPLING_RATE", 16000)
            // << hark::Input("STRIP_ZERO", true)
            // << hark::Input("ZERO_MEAN", false)
            << hark::Input("LEVEL_THRESHOLD", 60.f)
            << hark::Input("ZEROCROSS_THRESHOLD", 80)
            << hark::Input("HEAD_MARGIN", 400.f)
            << hark::Input("TAIL_MARGIN", 600.f);

        node_extracting
            << hark::Input("CHANNEL_WAV", node_vad_zc["OUTPUT"]);
        // << hark::Input("LENGTH", 512)
        // << hark::Input("SAMPLING_RATE", 16000);

        node_spectral_mean_normalization_incremental
            << hark::Input("INPUT", node_extracting["OUTPUT"])
            << hark::Input("NOT_EOF", true) // ToDo
            << hark::Input("SM_HISTORY", false) // ToDo
            // << hark::Input("FBANK_COUNT", 13)
            << hark::Input("PERIOD", 1)
            // << hark::Input("SM_ALGORITHM", "INCREMENTAL")
            // << hark::Input("SM_FILENAME", "")
            // << hark::Input("SM_HISTORY_FILENAME", "")
            // << hark::Input("IGNORE_FRAMES", false)
            << hark::Input("BASENAME", "")
            << hark::Input("OUTPUT_FNAME", "");
            // << hark::Input("SM_EXPORT_FILENAME", "")
            // << hark::Input("SM_EXPORT_ALGORITHM", "LAST_SRC");

        output["OUTPUT"] << node_spectral_mean_normalization_incremental["OUTPUT"];
    }
};

class Subscriber: public hark::node::SubscribeData
{
    void
    receive(data output)
    {
        std::stringstream ss;
        ss << ">>>> received: " << output << std::endl;
        std::cout << ss.str();
    }
};

class Main: public hark::NetworkDef
{
public:
    void
    build(hark::Network &network, hark::DataSourceMap &input, hark::DataSinkMap &output){

        //auto &pub = network.create<hark::Node, hark::node::PublishData, hark::dispatch::Repeat>("Publisher");
        //auto &sub = network.create<hark::Node, Subscriber>("Subscriber");
        //auto &loop = network.create<hark::Network, Recording>("Recording");

        //loop << hark::Input("INPUT", pub["OUTPUT"]);
        //sub  << hark::Input("INPUT", loop["OUTPUT"]);

        auto &pub = network.create_node(new hark::node::PublishData(), new hark::dispatch::Repeat(), "Publisher");
        auto &sub = network.create_node(new Subscriber(), new hark::dispatch::TriggeredRepeat(), "Subscriber");
        auto &loop = network.create_network(new Recording(), new hark::dispatch::TriggeredRepeat(), "Recording")
            << hark::Input("INPUT", pub["OUTPUT"]);
        sub.add_input(hark::Input("INPUT", loop["OUTPUT"]));
    }

};

int
process()
{
    // hark:data data(123.45);
    // std::cout << data.as<int>() << std::endl;
    // std::cout << data.as<float>() << std::endl;
    //
    std::shared_ptr<Network> network(hark::Network::create<Main>("Main1"));

    auto worker = [](std::shared_ptr<Network> network){
        network->execute();
    };
    std::thread worker_thread(worker, network);

    {
        auto publisher = dynamic_cast<hark::node::PublishData *>(network->query_node("Publisher")->get_nodedef());
        std::default_random_engine engine;
        std::uniform_int_distribution<int16_t> dist(-4096, 4096);

        for(size_t i = 0; i < 1000; i++){
            std::vector<int16_t> v(512);
            int16_t m = 0;
            if((i / 200) % 2 == 1){
                m = 0;
                for(size_t j = 0; j < v.size(); j++){
                    int16_t x = dist(engine);
                    v[j] = x;
                    if(m < x){
                        m = x;
                    }
                    if(m < -x){
                        m = -x;
                    }
                }
            }else{
                for(size_t j = 0; j < v.size(); j++){
                    v[j] = 0;
                }
                m = 0;
            }

            {
                std::stringstream ss;
                ss << "<<<< send: count=" << i << ", max=" << m << std::endl;
                std::cout << ss.str();
            }
            
            publisher->push(hark::data(v));
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }

        publisher->close();
    }

    network->stop();
    worker_thread.join();

    return 0;
}

int
main(int argc, char *argv[])
{
    try{
        process();
    }catch(std::exception &ex){
        std::cerr << "Error in process: " << ex.what() << std::endl;
        return EXIT_FAILURE;
    }catch(...){
        std::cerr << "Error in process: " << "not std::exception type" << std::endl;
        throw;
    }
    return EXIT_SUCCESS;
}
